
/**
  * Copyright (c) 2009, 2011 Alexander Churanov. All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  * notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above
  * copyright notice, this list of conditions and the following
  * disclaimer in the documentation and/or other materials provided
  * with the distribution.
  *
  * THIS SOFTWARE IS PROVIDED BY ALEXANDER CHURANOV ``AS IS'' AND ANY
  * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ALEXANDER CHURANOV BE
  * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
  * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */

#ifndef _CH_EXERCISIX_HH_INCLUDED_
#define _CH_EXERCISIX_HH_INCLUDED_

#include <cmath>
#include <cstdlib>

#include <algorithm>
#include <iomanip>
#include <iostream>
#include <memory>   // for std::auto_ptr
#include <sstream>
#include <stdexcept>
#include <vector>

#if !defined(__clang__) && (defined(__GLIBCXX__) || defined(__GLIBCPP__))
#  include <cxxabi.h> // demangling on gcc
#endif

namespace exercisix
{
  typedef void (*test_fun)();
  
  struct test_case
  {
    test_case(
      int line,
      test_fun fun,
      std::string const& descr,
      std::string const& filename)
      
      : line(line),
	fun(fun),
	descr(descr),
	filename(filename)
    { }
    
    static bool less_by_line(test_case const& lhs,
			     test_case const& rhs)
    {
      return lhs.line < rhs.line;
    }
    
    int line;
    test_fun fun;
    std::string descr;
    std::string filename;
  };
  
  class failure
  {
  public:
    failure()
      : test(0)
    { }
    
    virtual ~failure() { }
    virtual void report() = 0;
    
    void set_test(test_case const* test)
    {
      this->test = test;
    }
    
    struct same_test_case
    {
      same_test_case(test_case const* test)
	: test(test)
      { }
      
      bool const operator()(failure const* item) const
      {
	return item->test == this->test;
      }
      
      test_case const* test;
    };
    
  protected:
    test_case const* test;
  };
  
  class comparison_failure
    : public failure
  {
  public:
    comparison_failure(int line,
		       std::string const& op,
		       std::string const& arg1,
		       std::string const& arg2,
		       std::string const& value1,
		       std::string const& value2)
      
      : line(line), op(op),
	arg1(arg1), arg2(arg2),
	value1(value1), value2(value2)
    { }
    
  private:
    void report()
    {
      using namespace std;
      
      cout << test->filename << ":" << line << ":expectation failed" << endl
           << endl
           << test->descr << endl
           << endl
           << "operand1: " << arg1   << endl
           << "operand2: " << arg2   << endl
           << "operator: " << op     << endl
           << "value1  : "   << value1 << endl
           << "value2  : "   << value2 << endl;
    }
    
    int line;
    std::string op;
    std::string arg1;
    std::string arg2;
    std::string value1;
    std::string value2;
  };
  
  class exception_failure
    : public failure
  {
  public:
    exception_failure(int line,
		      std::string const& expected,
		      std::string const& encountered)
      
      : line(line),
	expected(expected), encountered(encountered)
    { }
    
  private:
    void report()
    {
      using namespace std;
      
      cout << test->filename << ":"
	   << line << ":"
	   << "expectation failed" << endl;
      cout << endl << test->descr << endl;
      
      cout << endl;
      cout << "expected: " << expected << endl;
      cout << "encountered: " << encountered << endl;
    }
    
    int line;
    std::string expected;
    std::string encountered;
  };
  
  class unexpected_exception_failure
    : public failure
  {
  public:
    unexpected_exception_failure(
      int line,
      std::string const& type,
      std::string const& text)
      
      : line(line), type(type), text(text)
    { }
    
  private:
    void report()
    {
      using namespace std;
      
      cout << test->filename << ":" << line << ":"
	   << "unexpected exception" << endl
           << endl
           << test->descr << endl
           << endl
           << "exception type: " << type << endl
           << "message: " << text << endl;
    }
    
    int line;
    std::string type;
    std::string text;
  };
  
  class explicit_failure
    : public failure
  {
  public:
    explicit explicit_failure(int line)
      : line(line)
    { }
    
  private:
    void report()
    {
      using namespace std;
      
      cout << test->filename << ":"
	   << line << ":"
	   << "EXERCISIX_FAIL reached" << endl;
      cout << endl << test->descr << endl;
    }
    
    int line;
  };

#if !defined(__clang__) && (defined(__GLIBCXX__) || defined(__GLIBCPP__))
  std::string demangle(std::string const& name)
  {
    int status;
    return abi::__cxa_demangle(name.c_str(), 0, 0, &status);
  }
#else
  std::string demangle(std::string const& name)
  {
    return name; // no demangling
  }
#endif
  
  class test_suite
  {
  public:
    test_suite()
      : verbose_mode(false)
    { }
    
    ~test_suite()
    {
      failure_container_t::iterator it = failures.begin();
      
      while (it != failures.end())
      {
	delete (*it++);
      }
    }
    
    int run()
    {
      int i = 1;
      test_container_t::const_iterator it = tests.begin();
      
      for (; it != tests.end() && failures.empty(); ++it, ++i)
      {
	current = (&*it);
	
	print_before_message(*it, i);
	run_test_case(*it);
	print_after_message(*it);
      }
      
      if (failures.empty())
      {
	report_success();
	return EXIT_SUCCESS;
      }
      else {
	failures.front()->report();
	return EXIT_FAILURE;
      }
    }
    
    bool add_test(
      int line,
      test_fun fun,
      char const descr[],
      char const filename[])
    {
      tests.push_back(test_case(line, fun, descr, filename));
      return true;
    }
    
    void register_failure(std::auto_ptr <failure> error)
    {
      error->set_test(current);
      failures.push_back(error.get());
      error.release();
    }
    
    bool rename(char const name[])
    {
      this->name = name;
      return true;
    }
    
    void set_executable_name(char const name[])
    {
      this->executable_name = name;
      
      std::string::size_type const slash_idx = executable_name.find_last_of('/');
      
      if (slash_idx != std::string::npos)
        this->executable_name = this->executable_name.substr(slash_idx + 1);
    }
    
    void print_cases() const
    {
      using namespace std;
      
      int const max_line =
	max_element(tests.begin(), tests.end(),
		    test_case::less_by_line)->line;
      
      int i = 1;
      test_container_t::const_iterator it = tests.begin();
      
      while (it != tests.end())
      {
	cout << i << ": "
	     << "test"
	     << setw(std::log10(max_line) + 1)
	     << left << it->line << ":" << it->descr << endl;
	++it; ++i;
      }
    }
    
    void verbose(bool verbose_mode)
    {
      this->verbose_mode = verbose_mode;
    }
    
  private:
    std::string name;
    std::string executable_name;
    
    bool verbose_mode;
    
    typedef std::vector <test_case> test_container_t;
    test_container_t tests;
    
    typedef std::vector <failure*> failure_container_t;
    failure_container_t failures;
    
    test_case const* current;
    
  private:
    void run_test_case(test_case const& test)
    {
      try {
	test.fun();
      }
      catch (std::exception const& e)					
      {
	std::auto_ptr <failure> error(
	  new unexpected_exception_failure(
	    test.line, demangle(typeid(e).name()), e.what())
	  );
	register_failure(error);
      }
      catch(...)
      {
	std::auto_ptr <failure> error(
	  new unexpected_exception_failure(
	    test.line, "unknown type", "can't retrieve message")
	  );
	register_failure(error);
      }
    }
    
    void print_before_message(test_case const& test, int i)
    {
      if (verbose_mode)
	std::cout << i << ": " << test.descr << "...";
    }
    
    void print_after_message(test_case const& test)
    {
      if (verbose_mode)
      {
	bool const failed = find_if(
	  failures.begin(), failures.end(),
	  failure::same_test_case(&test)) != failures.end();
	
	std::cout << (failed ? "" : "SUCCESS") << std::endl;
      }
    }
    
    std::string const translation_unit_name()
    {
      using namespace std;
      
      if (tests.empty())
        return executable_name; // not helpful for navigating through
                                // errors, but empty suite should not
                                // have any
      
      string const& filename = tests.front().filename;
      
      string::size_type const dot = filename.find_last_of('.');
      
      return (dot == string::npos) ?
        filename : filename.substr(0, dot);
    }
    
    void report_success()
    {
      using namespace std;
      cout << (name.empty() ? translation_unit_name() : name)
		<< ": SUCCESS" << endl;
    }
  };
  
  test_suite suite;
  
  struct any
  {
    template <typename T>
    any(T const&)
    { }
  };
  
  std::ostream&
  operator<<(std::ostream& out, any const& value)
  {
    std::operator<<(out, "value can not be printed");
    return out;
  }
  
  void print_version()
  {
    std::cout << "This is exercisix version 1.2" << std::endl;
  }
  
} // namespace exercisix

int main(int argc, char* argv[])
{
  using namespace std;
  
  bool const list_mode =
    (find(argv, argv + argc, string("--list")) != argv + argc) or
    (find(argv, argv + argc, string("-l"))     != argv + argc);
  
  bool const verbose_mode =
    (find(argv, argv + argc, string("--verbose")) != argv + argc) or
    (find(argv, argv + argc, string("-v"))        != argv + argc);
  
  bool const version_mode =
    find(argv, argv + argc, string("--version")) != argv + argc;
  
  if (version_mode)
  {
    ::exercisix::print_version();
    return EXIT_SUCCESS;
  }
  
  if (list_mode)
  {
    ::exercisix::suite.print_cases();
    return EXIT_SUCCESS;
  }
  
  if (verbose_mode)
    ::exercisix::suite.verbose(true);
  
  ::exercisix::suite.set_executable_name(argv[0]);
  return ::exercisix::suite.run();
}

#define EXERCISIX_CONCAT_2( a, b ) a##b
#define EXERCISIX_CONCAT( a, b ) EXERCISIX_CONCAT_2( a, b )


#define TEST_SUITE(name)						\
  namespace exercisix							\
  {									\
    bool const EXERCISIX_CONCAT(renaming, __LINE__) =			\
      ::exercisix::suite.rename((name));				\
  } /* namespace exercisix */						\
									\
  namespace EXERCISIX_CONCAT(exercisix_test_suite_namespace, __LINE__)


#define TEST(descr)						 \
  void EXERCISIX_CONCAT(test, __LINE__)();			 \
								 \
  namespace exercisix						 \
  {								 \
    bool const EXERCISIX_CONCAT(registration_result, __LINE__) = \
      ::exercisix::suite.add_test(				 \
        __LINE__,			                         \
	EXERCISIX_CONCAT(test, __LINE__),			 \
	descr, __FILE__);					 \
  } /* namespace exercisix */					 \
								 \
  void EXERCISIX_CONCAT(test, __LINE__)()

namespace exercisix
{
  template <typename T>
  T pass_value(T value, std::string& s)
  {
    using namespace exercisix; /* for operator<< */
    
    std::ostringstream out;
    out << value;
    s = out.str();
    return value;
  }
  
} // namespace exercisix

#define EXPECT(arg1, op, arg2)                              \
                                                            \
  {                                                         \
    std::string value1_str;                                 \
    std::string value2_str;                                 \
                                                            \
    if (!(::exercisix::pass_value(arg1, value1_str)         \
          op                                                \
          ::exercisix::pass_value(arg2, value2_str)))       \
    {                                                       \
      std::auto_ptr < ::exercisix::failure >                \
        exercisix_error(                                    \
          new ::exercisix::comparison_failure(              \
            __LINE__, #op, #arg1, #arg2,                    \
            value1_str, value2_str                          \
            )                                               \
          );                                                \
                                                            \
      ::exercisix::suite.register_failure(exercisix_error); \
    }                                                       \
  }

#define EXPECT_NO_EXCEPTIONS(expr)					\
  try {									\
    (expr);								\
  }									\
  catch (std::exception const& exercisix_e##__LINE__)			\
  {									\
    std::auto_ptr < ::exercisix::failure >				\
      exercisix_error##__LINE__(					\
      new ::exercisix::exception_failure(				\
	__LINE__, "no exceptions",					\
	std::string("exception of type \"") +				\
	::exercisix::demangle(typeid(exercisix_e##__LINE__).name()) +   \
        "\"")                                                           \
      );								\
    ::exercisix::suite.register_failure(exercisix_error##__LINE__);	\
  }									\
  catch(...)								\
  {									\
    std::auto_ptr < ::exercisix::failure >				\
      exercisix_error##__LINE__(					\
      new ::exercisix::exception_failure(				\
	__LINE__, "no exceptions", "exception of unknown type")		\
      );								\
    ::exercisix::suite.register_failure(exercisix_error##__LINE__);	\
  }

#define EXPECT_EXCEPTION(expr, type)					\
  {									\
    bool exercisix_thrown##__LINE__ = false;				\
									\
    std::string exercisix_expected##__LINE__(				\
      std::string("exception of type \"") +				\
      ::exercisix::demangle(typeid(type).name()) + "\""                 \
      );  								\
    									\
    try {								\
      (expr);								\
    }									\
    catch (type const& exercisix_e##__LINE__)				\
    {									\
      exercisix_thrown##__LINE__ = true;				\
    }									\
    catch (std::exception const& exercisix_e##__LINE__)			\
    {									\
      std::auto_ptr < ::exercisix::failure >				\
	exercisix_error##__LINE__(					\
	new ::exercisix::exception_failure(				\
	  __LINE__, exercisix_expected##__LINE__,			\
	  std::string("exception of type \"") +				\
	  ::exercisix::demangle(typeid(exercisix_e##__LINE__).name())   \
          + "\"")                                                       \
	);								\
      ::exercisix::suite.register_failure(exercisix_error##__LINE__);	\
									\
      exercisix_thrown##__LINE__ = true;				\
    }									\
    catch(...)								\
    {									\
      std::auto_ptr < ::exercisix::failure >				\
	exercisix_error##__LINE__(					\
	new ::exercisix::exception_failure(				\
	  __LINE__, exercisix_expected##__LINE__,			\
	  "exception of unknown type")					\
	);								\
      ::exercisix::suite.register_failure(exercisix_error##__LINE__);	\
									\
      exercisix_thrown##__LINE__ = true;				\
    }									\
    									\
    if (!exercisix_thrown##__LINE__)					\
    {									\
      std::auto_ptr < ::exercisix::failure >				\
	exercisix_error##__LINE__(					\
	new ::exercisix::exception_failure(				\
	  __LINE__, exercisix_expected##__LINE__,			\
	  "no exceptions at all")					\
	);								\
      ::exercisix::suite.register_failure(exercisix_error##__LINE__);	\
    }									\
  }

#define EXERCISIX_FAIL							\
  {									\
    std::auto_ptr < ::exercisix::failure >				\
      exercisix_error##__LINE__(					\
	new ::exercisix::explicit_failure(__LINE__)			\
	);								\
    ::exercisix::suite.register_failure(exercisix_error##__LINE__);	\
  }

#endif // compiler guard
